"""
This module provides frequency arithmetic for linear and quadratic responses.
"""

from enum import Enum
import numpy


class Subset(Enum):
    """
    Description of the subsets of interactive frequencies.
    """

    LINEAR = 0x000F
    """Linear frequencies :math:`f_i`"""
    DIAGONAL = 0x00F0
    """Diagonal (harmonic) frequencies :math:`2 f_i`"""
    PLUS = 0x0F00
    """Sum frequencies :math:`f_i + f_j`"""
    MINUS = 0xF000
    """Difference frequencies :math:`|f_i - f_j|`"""
    ALL = 0xFFFF
    """All frequencies, equivalent to LINEAR, QUADRATIC"""
    QUADRATIC = 0xFFF0
    """Quadratic frequencies, equivalent to DIAGONAL, PLUS, MINUS"""


def hertz2fick(dt, duration, frequency):
    """
    Convert a frequency given in hertz to the corresponding FFT index.
    Note that negative frequencies are converted to the right of positive frequencies.

    :param dt: sampling time (seconds)
    :type dt: float
    :param duration: experiment duration (seconds) such that :math:`1 / duration` represents the minimal frequency
    :type duration: float
    :param frequency: frequency (hertz) to convert
    :type frequency: float
    :return: index of the frequency in the FFT
    :rtype: float

    .. highlight:: python
    .. code-block:: python

        >>> qsa.domain.hertz2fick(0.001, 10, 1)
        10
        >>> qsa.domain.hertz2fick(0.001, 10, -1)
        9990
    """
    n = duration / dt if frequency < 0 else 0
    return int(round(n + frequency * duration))


def hertz2ficks(dt, duration, frequencies):
    """
    Convert an array of frequencies in hertz to the corresponding array of FFT indices.

    :param dt: sampling time (seconds)
    :type dt: float
    :param duration: experiment duration (seconds) such that :math:`1 / duration` represents the minimal frequency
    :type duration: float
    :param frequencies: array of frequencies (hertz) to convert
    :type frequencies: numpy.ndarray
    :return: indices of the frequencies in the FFT
    :rtype: numpy.ndarray

    .. highlight:: python
    .. code-block:: python

        >>> qsa.domain.hertz2ficks(0.001, 10, numpy.array([-10, -1, +1, +10]))
        [9900, 9990, 10, 100]
    """
    return list(map(
        lambda frequency: hertz2fick(dt, duration, frequency),
        frequencies))


def order(frequencies, symmetric=True, subset=Subset.ALL):
    """
    Compute first and second order interactive frequencies.

    :param frequencies: array of fundamental frequencies (hertz) ; they must be strictly positive
    :type frequencies: numpy.ndarray
    :param symmetric: True to generate both negative and positive frequencies ; False to generate positive frequencies only
    :type symmetric: bool
    :param subset: specify the frequency combinations (see also :py:class:`qsa.domain.Subset`)
    :type subset: enumerate
    :return: interactive frequencies (hertz) generated from fundamental frequencies
    :rtype: numpy.ndarray

    .. highlight:: python
    .. code-block:: python

        >>> qsa.domain.order(numpy.array([2, 7]), True, qsa.domain.Subset.QUADRATIC)
        array([-14.,  -9.,  -5.,  -4.,   4.,   5.,   9.,  14.])
    """
    mixing = numpy.array([])
    for i in range(0, len(frequencies)):
        fi = frequencies[i]
        if Subset.LINEAR.value & subset.value:
            mixing = numpy.append(mixing, fi)
        if Subset.DIAGONAL.value & subset.value:
            mixing = numpy.append(mixing, 2 * fi)
        for j in range(i + 1, len(frequencies)):
            fj = frequencies[j]
            if Subset.PLUS.value & subset.value:
                mixing = numpy.append(mixing, fi + fj)
            if Subset.MINUS.value & subset.value:
                mixing = numpy.append(mixing, abs(fi - fj))
    result = mixing
    if symmetric:
        negative = numpy.negative(result)
        result = numpy.concatenate([negative, result])
    return numpy.sort(result)



"""
This module computes linear and quadratic analyses of measurements.
"""

import numpy

import qsa.domain
import qsa.measurement


class Analysis:
    """
    Computation of linear and quadratic analysis of a measurement.
    """

    def __init__(self, measurement):
        """
        Constructor.

        :param measurement: the measurement to be analyzed
        :type measurement: qsa.measurement.Measurement
        """
        self.__dt = measurement.dt
        self.__duration = measurement.duration
        self.__frequencies = numpy.copy(measurement.frequencies)
        self.__t = numpy.copy(measurement.t)
        self.__x = numpy.copy(measurement.x)
        self.__y = numpy.copy(measurement.y)
        #
        self.__x_sp = spectralize(self.__x)
        self.__y_sp = spectralize(self.__y)
        #
        self.__x_dc = numpy.real(self.__x_sp[0])
        self.__y_dc = numpy.real(self.__y_sp[0])
        #
        self.__frequencies1 = qsa.domain.order(
            self.__frequencies,
            symmetric=False,
            subset=qsa.domain.Subset.LINEAR)
        self.__ficks1 = qsa.domain.hertz2ficks(
            self.__dt, self.__duration, self.__frequencies1)
        self.__x_sp1 = self.__x_sp[self.__ficks1]
        self.__x1 = self.__approximate(self.__x_sp, qsa.domain.Subset.LINEAR)
        self.__y_sp1 = self.__y_sp[self.__ficks1]
        self.__y1 = self.__approximate(self.__y_sp, qsa.domain.Subset.LINEAR)
        #
        self.__frequencies2 = qsa.domain.order(
            self.__frequencies,
            symmetric=False,
            subset=qsa.domain.Subset.QUADRATIC)
        self.__ficks2 = qsa.domain.hertz2ficks(
            self.__dt, self.__duration, self.__frequencies2)
        self.__y_sp2 = self.__y_sp[self.__ficks2]
        self.__y2 = self.__approximate(self.__y_sp, qsa.domain.Subset.QUADRATIC)
        #
        self.__frequencies12 = qsa.domain.order(
            self.__frequencies,
            symmetric=False,
            subset=qsa.domain.Subset.ALL)
        self.__ficks12 = qsa.domain.hertz2ficks(
            self.__dt, self.__duration, self.__frequencies12)
        self.__y_sp12 = self.__y_sp[self.__ficks12]
        self.__y12 = self.__approximate(self.__y_sp, qsa.domain.Subset.ALL)
        #
        self.__frequencies2d = qsa.domain.order(
            self.__frequencies,
            symmetric=False,
            subset=qsa.domain.Subset.DIAGONAL)
        self.__ficks2d = qsa.domain.hertz2ficks(
            self.__dt, self.__duration, self.__frequencies2d)
        self.__y_sp2d = self.__y_sp[self.ficks2d]
        self.__frequencies2p = qsa.domain.order(
            self.__frequencies,
            symmetric=False,
            subset=qsa.domain.Subset.PLUS)
        self.__ficks2p = qsa.domain.hertz2ficks(
            self.__dt, self.__duration, self.__frequencies2p)
        self.__y_sp2p = self.__y_sp[self.__ficks2p]
        self.__frequencies2m = qsa.domain.order(
            self.__frequencies,
            symmetric=False,
            subset=qsa.domain.Subset.MINUS)
        self.__ficks2m = qsa.domain.hertz2ficks(
            self.__dt, self.__duration, self.__frequencies2m)
        self.__y_sp2m = self.__y_sp[self.__ficks2m]
        #
        self.__linear = self.__y_sp1 / self.__x_sp1
        self.__quadratic = self.__qmatrix()
        self.__eigen = numpy.linalg.eigvalsh(self.__quadratic)

    @property
    def dt(self):
        """
        Get sampling time.

        :return: dt (seconds)
        :rtype: float
        """
        return self.__dt

    @property
    def duration(self):
        """
        Get experiment duration.

        :return: duration (seconds)
        :rtype: float
        """
        return self.__duration

    @property
    def frequencies(self):
        """
        Get fundamental frequencies.

        :return: copy of the array of fundamental frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies)

    @property
    def t(self):
        """
        Get time.

        :return: copy of the time vector (seconds)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__t)

    @property
    def x(self):
        """
        Get stimulation.

        :return: copy of the stimulation vector
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__x)

    @property
    def y(self):
        """
        Get response.

        :return: copy of the response vector
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y)

    @property
    def x_sp(self):
        """
        Get stimulation spectrum.

        :return: copy of the stimulation spectrum vector
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__x_sp)

    @property
    def y_sp(self):
        """
        Get response spectrum.

        :return: copy of the response spectrum vector
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp)

    @property
    def x_dc(self):
        """
        Get stimulation DC.

        :return: stimulation DC
        :rtype: float
        """
        return self.__x_dc

    @property
    def y_dc(self):
        """
        Get response DC.

        :return: response DC
        :rtype: float
        """
        return self.__y_dc

    @property
    def frequencies1(self):
        """
        Get positive linear frequencies.

        :return: copy of the array of positive linear frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies1)

    @property
    def ficks1(self):
        """
        Get FFT indices of frequencies1.

        :return: copy of the array of FFT indices of frequencies1
        :rtype: numpy.ndarray
        """
        return self.copy(self.__ficks1)

    @property
    def x_sp1(self):
        """
        Get stimulation spectrum over ficks1.

        :return: copy of the stimulation spectrum over ficks1
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__x_sp1)

    @property
    def x1(self):
        """
        Get linear approximation of stimulation

        :return: copy of the linear approximation of stimulation
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__x1)

    @property
    def y_sp1(self):
        """
        Get response spectrum over ficks1.

        :return: copy of the response spectrum over ficks1
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp1)

    @property
    def y1(self):
        """
        Get dc + linear approximation of response

        :return: copy of the dc + linear approximation of response
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y1)

    @property
    def frequencies2(self):
        """
        Get positive quadratic frequencies.

        :return: copy of the array of positive quadratic frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies2)

    @property
    def ficks2(self):
        """
        Get FFT indices of frequencies2.

        :return: copy of the array of FFT indices of frequencies2
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__ficks2)

    @property
    def y_sp2(self):
        """
        Get response spectrum over ficks2.

        :return: copy of the response spectrum over ficks2
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp2)

    @property
    def y2(self):
        """
        Get dc + quadratic approximation of response

        :return: copy of the dc + quadratic approximation of response
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y2)

    @property
    def frequencies12(self):
        """
        Get positive linear + quadratic frequencies.

        :return: copy of the array of positive linear + quadratic frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies12)

    @property
    def ficks12(self):
        """
        Get FFT indices of frequencies12.

        :return: copy of the array of FFT indices of frequencies12
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__ficks12)

    @property
    def y_sp12(self):
        """
        Get response spectrum over ficks12.

        :return: copy of the response spectrum over ficks12
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp12)

    @property
    def y12(self):
        """
        Get dc + linear + quadratic approximation of response

        :return: copy of the dc + linear + quadratic approximation of response
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y12)

    @property
    def frequencies2d(self):
        """
        Get positive diagonal frequencies.

        :return: copy of the array of positive diagonal frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies2d)

    @property
    def ficks2d(self):
        """
        Get FFT indices of frequencies2d.

        :return: copy of the array of FFT indices of frequencies2d
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__ficks2d)

    @property
    def y_sp2d(self):
        """
        Get response spectrum over ficks2d.

        :return: copy of the response spectrum over ficks2d
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp2d)

    @property
    def frequencies2p(self):
        """
        Get positive plus frequencies.

        :return: copy of the array of positive plus frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies2p)

    @property
    def ficks2p(self):
        """
        Get FFT indices of frequencies2p.

        :return: copy of the array of FFT indices of frequencies2p
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__ficks2p)

    @property
    def y_sp2p(self):
        """
        Get response spectrum over ficks2p.

        :return: copy of the response spectrum over ficks2p
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp2p)

    @property
    def frequencies2m(self):
        """
        Get positive minus frequencies.

        :return: copy of the array of positive minus frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies2m)

    @property
    def ficks2m(self):
        """
        Get FFT indices of frequencies2m.

        :return: copy of the array of FFT indices of frequencies2m
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__ficks2m)

    @property
    def y_sp2m(self):
        """
        Get response spectrum over ficks2m.

        :return: copy of the response spectrum over ficks2m
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y_sp2m)

    @property
    def linear(self):
        """
        Get linear transfer function over ficks1.

        :return: copy of the transfer function over ficks1
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__linear)

    @property
    def quadratic(self):
        """
        Get quadratic transfer function, which is the QSA matrix.

        :return: copy of the  quadratic transfer function
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__quadratic)

    @property
    def eigen(self):
        """
        Get eigenvalues of the QSA matrix.

        :return: copy of the eigenvalues
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__eigen)

    def __bcoefficient(self, f, i, j):
        if f[i] == - f[j]:
            return 0
        k = 1 if i == j else 0.5
        fi = qsa.domain.hertz2fick(self.dt, self.duration, f[i])
        fj = qsa.domain.hertz2fick(self.dt, self.duration, f[j])
        fij = qsa.domain.hertz2fick(self.dt, self.duration, f[i] + f[j])
        rij = self.y_sp[fij]
        si = self.x_sp[fi]
        sj = self.x_sp[fj]
        return k * rij / (si * sj)

    def __bmatrix(self):
        n = len(self.frequencies)
        b = numpy.zeros([2 * n, 2 * n], dtype=complex)
        f = qsa.domain.order(
            self.frequencies,
            symmetric=True,
            subset=qsa.domain.Subset.LINEAR)
        for i in range(0, 2 * n):
            for j in range(0, 2 * n):
                b[i, j] = self.__bcoefficient(f, i, j)
        return b

    def __qmatrix(self):
        return numpy.flipud(self.__bmatrix())

    def __approximate(self, z_sp, subset):
        frequencies_generated = qsa.domain.order(
            self.frequencies,
            symmetric=True,
            subset=subset)
        ficks = qsa.domain.hertz2ficks(
            self.dt,
            self.duration,
            frequencies_generated)
        za_sp = numpy.zeros(numpy.size(z_sp), dtype='complex')
        za_sp[0] = z_sp[0]
        za_sp[ficks] = z_sp[ficks]
        return unspectralize(za_sp)


def spectralize(z):
    """
    Compute discrete Fourier transform of a real signal.
    The FFT is normalized by the length of the signal. In particular,
    :math:`cos(t)` is transformed to a sum of two complex exponentials with half
    amplitude

    .. math::

        cos(t) = \\frac{1}{2} e^{i t} + \\frac{1}{2} e^{- i t}

    :param z: signal as a time series of real values
    :type z: numpy.ndarray
    :return: array of Fourier coefficients (complex numbers)
    :rtype: numpy.ndarray

    .. highlight:: python
    .. code-block:: python

        >>> t = numpy.linspace(0, 10, 10000)
        >>> z = numpy.cos(2 * numpy.pi * t)
        >>> z_sp = qsa.analysis.spectralize(z)
        >>> numpy.max(numpy.abs(z_sp))
        0.5000241787423103
    """
    return numpy.fft.fft(z) / len(z)


def unspectralize(z_sp):
    """
    Compute inverse of a Fourier transform obtained with :py:func:`spectralize`.
    By definition, unspectralize(spectralize(z)) equals z.

    :param z_sp: a Fourier transform obtained with :py:func:`spectralize`
    :type z_sp: numpy.ndarray
    :return: a real signal for which Fourier transform is z_sp
    :rtype: numpy.ndarray

    .. highlight:: python
    .. code-block:: python

        >>> t = numpy.linspace(0, 10, 10000)
        >>> z = numpy.cos(2 * numpy.pi * t)
        >>> z_sp = qsa.analysis.spectralize(z)
        >>> z_sp_inv = qsa.analysis.unspectralize(z_sp)
        >>> numpy.max(z - z_sp_inv)
        7.771561172376096e-16
    """
    z_sp_pure = numpy.copy(z_sp)
    z_sp_pure[0] = 0
    return numpy.real(numpy.fft.ifft(z_sp_pure) * len(z_sp) + z_sp[0])



"""
This module represents an RTXI experiment as an array of measurements (see also :py:class:`qsa.measurement.Measurement`).
"""

import json
import numpy

import qsa.measurement


class Experiment:
    """
    Representation of an RTXI experiment as an array of measurements.
    Each measurement corresponds to the multisinusoidal part of a single trace.
    """

    def __init__(self, filename):
        """
        Constructor.

        :param filename: Filename of the JSON file generated by the RTXI program.
        :type filename: str
        """
        with open(filename, 'r') as file:
            text = file.read()
        j = json.loads(text)
        self.__version = j['version']
        self.__dt = j['dt']
        self.__duration = j['duration']
        self.__frequencies = numpy.array(j['frequencies'])
        self.__amplitudes = numpy.array(j['amplitudes'])
        self.__phases = numpy.array(j['phases'])
        self.__rest_level = j['rest_level']
        self.__step_level = j['step_level']
        self.__step_delay = j['step_delay']
        self.__drop_delay = j['drop_delay']
        self.__trace_count = j['trace_count']
        self.__trace_pause = j['trace_pause']
        self.__trace_alternance = j['trace_alternance']
        self.__traces = []
        for i in range(0, len(j['traces'])):
            def trace(): return None

            trace.step = lambda: None
            trace.step.time = numpy.array(j['traces'][i]['step']['time'])
            trace.step.stimulation = numpy.array(
                j['traces'][i]['step']['stimulation'])
            trace.step.response = numpy.array(
                j['traces'][i]['step']['response'])
            trace.multisine = lambda: None
            trace.multisine.time = numpy.array(
                j['traces'][i]['multisine']['time'])
            trace.multisine.stimulation = numpy.array(
                j['traces'][i]['multisine']['stimulation'])
            trace.multisine.response = numpy.array(
                j['traces'][i]['multisine']['response'])
            trace.drop = lambda: None
            trace.drop.time = numpy.array(j['traces'][i]['drop']['time'])
            trace.drop.stimulation = numpy.array(
                j['traces'][i]['drop']['stimulation'])
            trace.drop.response = numpy.array(
                j['traces'][i]['drop']['response'])
            self.__traces.append(trace)

    @property
    def version(self):
        """
        Get version.

        :return: version
        :rtype: str
        """
        return self.__version

    def count(self):
        """
        Count measurements.

        :return: number of measurements
        :rtype: int
        """
        return len(self.__traces)

    def get_measurement(self, index):
        """
        Get measurement.

        :param index: index of measurement in array of measurements
        :type index: int
        :return: measurement at the specified index
        :rtype: qsa.measurement.Measurement
        :raises IndexError: if index is out of range
        """
        if index < 0 or index >= len(self.__traces):
            raise IndexError('Invalid measurement index (%d)' % index)
        return qsa.measurement.Measurement(
            self.__dt,
            self.__duration,
            self.__frequencies,
            self.__traces[index].multisine.time,
            self.__traces[index].multisine.stimulation,
            self.__traces[index].multisine.response)

    def average(self, indices):
        """
        Compute averaging of several measurements.

        :param indices: array of measurement indices to take into account in averaging
        :type indices: array
        :return: averaged measurement
        :rtype: qsa.measurement.Measurement
        :raises NotImplementedError: if trace_alternance < 0
        :raises IndexError: if index is out of range
        """
        if self.__trace_alternance < 0:
            raise NotImplementedError(
                'Averaging with trace alternance is not yet implemented')
        t = 0
        x = 0
        y = 0
        n = len(indices)
        for i in range(0, n):
            index = indices[i]
            if index < 0 or index >= len(self.__traces):
                raise IndexError('Invalid measurement index (%d)' % index)
            m = self.get_measurement(index)
            t = t + m.t / n
            x = x + m.x / n
            y = y + m.y / n
        return qsa.measurement.Measurement(
            self.__dt,
            self.__duration,
            self.__frequencies,
            t,
            x,
            y)



"""
This modules provides graphical tools to display analysis results.
All functions are based on the matplotlib plotting library.
"""

import numpy
import matplotlib.pyplot as plt

import qsa.domain


def ion():
    """
    Turn the interactive mode on.

    :return: result of :py:func:`matplotlib.pyplot.ion`
    """
    plt.ion()


def figure():
    """
    Create a new figure.

    :return: result of :py:func:`matplotlib.pyplot.figure`
    """
    plt.figure()


def subplot(nrows, ncols, index):
    """
    Add a subplot to the current figure.

    :param nrows: number of rows
    :type nrows: int
    :param ncols: number of columns
    :type ncols: int
    :param index: index of row and column ; starts at 1 in the upper left corner and increases to the right
    :type index: int
    :return: result of :py:func:`matplotlib.pyplot.subplot`
    """
    plt.subplot(nrows, ncols, index)


def plot_stimulation(a):
    """
    Plot the stimulation in the time and frequency domain.
    A new figure with subplots is automatically created.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :return: None
    """
    plt.figure()
    plt.subplot(2, 1, 1)
    plt.plot(a.t, a.x)
    plt.title('Stimulation')
    plt.xlabel('Time (s)')
    plt.ylabel('Input')
    plt.tight_layout()
    plt.subplot(2, 1, 2)
    plt.plot(numpy.abs(a.x_sp))
    plt.title('Stimulation (FFT)')
    plt.xlabel('Frequency (Hertz)')
    plt.ylabel('Input')
    plt.tight_layout()


def plot_response(a):
    """
        Plot the response in the time and frequency domain.
        A new figure with subplots is automatically created.

        :param a: QSA analysis
        :type a: qsa.analysis.Analysis
        :return: None
    """
    plt.figure()
    plt.subplot(2, 1, 1)
    plt.plot(a.t, a.y)
    plt.title('Response')
    plt.xlabel('Time (s)')
    plt.ylabel('Output')
    plt.tight_layout()
    plt.subplot(2, 1, 2)
    plt.plot(numpy.abs(a.y_sp))
    plt.title('Response (FFT)')
    plt.xlabel('Frequency (Hertz)')
    plt.ylabel('Output')
    plt.tight_layout()


def plot_comparison(a, legend=True, new_figure=True):
    """
    Plot the comparison between response, linear and linear + quadratic analyses in the time domain.
    The graphic is embedded in the current figure, or a new figure can be created.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :param legend: True (default) if legend should be displayed
    :type legend: bool
    :param new_figure: True (default) if a new figure should be created
    :type new_figure: bool
    :return: None
    """
    if new_figure:
        plt.figure()
    plt.plot(a.t, a.y, c='b')
    plt.plot(a.t, a.y1, c='g')
    plt.plot(a.t, a.y12, c='r')
    plt.title('Comparison', fontsize=16)
    plt.xlabel('Frequency (Hertz)')
    plt.ylabel('Responses')
    if legend:
        plt.legend(['Y', 'Y1', 'Y12'])
    plt.tight_layout()


def plot_pure(a, new_figure=True):
    """
    Plot the pure quadratic analysis in the time domain.
    The pure quadratic analysis has no dc and no linear component.
    The graphic is embedded in the current figure, or a new figure can be created.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :param new_figure: True (default) if a new figure should be created
    :type new_figure: bool
    :return: None
    """
    if new_figure:
        plt.figure()
    plt.plot(a.t, a.y2 - a.y_dc)
    plt.title('Pure quadratic')
    plt.xlabel('Frequency (Hertz)')
    plt.ylabel('Response Y2')
    plt.tight_layout()


def plot_transfer(a, new_figure=True):
    """
    Plot the linear transfer function in the frequency domain.
    The function represented is :math:`\\left|\\frac{Y(\\omega_k)}{X(\\omega_k)}\\right|`
    where :math:`\\omega_k` describe the finite set of positive linear frequencies.
    The graphic is embedded in the current figure, or a new figure can be created.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :param new_figure: True (default) if a new figure should be created
    :type new_figure: bool
    :return: None
    """
    if new_figure:
        plt.figure()
    plt.stem(a.frequencies1, numpy.abs(a.linear), linefmt='gray')
    plt.plot(a.frequencies1, numpy.abs(a.linear), c='b')
    plt.title('Linear transfer', fontsize=16)
    plt.xlabel('Frequency (Hertz)')
    plt.ylabel('Magnitude')
    plt.tight_layout()


def plot_flat(a, subset=qsa.domain.Subset.ALL, legend=True, new_figure=True):
    """
    Plot the response in the frequency domain.
    The function represented is :math:`\\left|Y(\\Omega_k)\\right|`
    where :math:`\\Omega_k` is a subset of positive interactive frequencies.
    The graphic is embedded in the current figure, or a new figure can be created.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :param subset: subset of interactive frequencies
    :type subset: qsa.domain.Subset
    :param legend: True (default) if legend should be displayed
    :type legend: bool
    :param new_figure: True (default) if a new figure should be created
    :type new_figure: bool
    :return: None
    """
    if new_figure:
        plt.figure()
    labels = []
    if qsa.domain.Subset.LINEAR.value & subset.value:
        plt.plot(a.frequencies1, numpy.abs(a.y_sp1), c='k')
        labels.append('Y1')
    if qsa.domain.Subset.DIAGONAL.value & subset.value:
        plt.scatter(a.frequencies2d, numpy.abs(a.y_sp2d), c='g')
        labels.append('Y2 diag')
    if qsa.domain.Subset.PLUS.value & subset.value:
        plt.scatter(a.frequencies2p, numpy.abs(a.y_sp2p), c='r')
        labels.append('Y2 plus')
    if qsa.domain.Subset.MINUS.value & subset.value:
        plt.scatter(a.frequencies2m, numpy.abs(a.y_sp2m), c='b')
        labels.append('Y2 minus')
    plt.title('Flat response', fontsize=16)
    plt.xlabel('Frequency (Hertz)')
    plt.ylabel('Magnitude')
    if legend:
        plt.legend(labels)
    plt.tight_layout()


def plot_qmatrix(a, new_figure=True):
    """
    Plot the quadratic transfer function in the frequency domain.
    The function represented is :math:`\\left|Q_{i,j}\\right|`
    where :math:`(Q_{i,j})` is the QSA matrix.
    The graphic is embedded in the current figure, or a new figure can be created.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :param new_figure: True (default) if a new figure should be created
    :type new_figure: bool
    :return: None
    """
    if new_figure:
        plt.figure()
    plt.matshow(numpy.abs(a.quadratic), cmap='jet', fignum=False)
    plt.colorbar()
    plt.title('QSA matrix', fontsize=16)
    gca = plt.gca()
    gca.set_xticklabels([])
    gca.set_yticklabels([])
    plt.tight_layout()


def plot_qeigen(a, new_figure=True):
    """
    Plot the eigenvalues of the QSA matrix.
    These are sorted by magnitude. Note that eigenvalues are real since the QSA matrix is Hermitian.

    :param a: QSA analysis
    :type a: qsa.analysis.Analysis
    :param new_figure: True (default) if a new figure should be created
    :type new_figure: bool
    :return: None
    """
    if new_figure:
        plt.figure()
    eigen = sorted(a.eigen, key=numpy.abs, reverse=True)
    plt.bar(numpy.arange(len(eigen)), eigen)
    plt.title('QSA eigenvalues')
    plt.tight_layout()



"""
This module represents a single trace measurement in the stimulation-response paradigm.
A measurement is applicable to both physical and virtual experiments.
"""

import numpy


class Measurement:
    """
    Implementation of a single trace measurement.
    """

    def __init__(self, dt, duration, frequencies, t, x, y):
        """
        Constructor.

        :param dt: sampling time (seconds)
        :type dt: float
        :param duration: experiment duration (seconds) such that :math:`1 / duration` represents the minimal frequency
        :type duration: float
        :param frequencies: array of fundamental frequencies (hertz)
        :type frequencies: numpy.ndarray
        :param t: time vector (seconds)
        :type t: numpy.ndarray
        :param x: stimulation vector
        :type x: numpy.ndarray
        :param y: response vector
        :type y: numpy.ndarray
        :raises ValueError: if t,x,y have wrong size
        """
        if len(t) != round(duration / dt):
            raise ValueError("Invalid length (%f) for t vector" % len(t))
        if len(x) != len(t):
            raise ValueError("Invalid length (%f) for x vector" % len(x))
        if len(y) != len(t):
            raise ValueError("Invalid length (%f) for y vector" % len(y))
        self.__dt = dt
        self.__duration = duration
        self.__frequencies = numpy.copy(frequencies)
        self.__t = numpy.copy(t)
        self.__x = numpy.copy(x)
        self.__y = numpy.copy(y)

    @property
    def dt(self):
        """
        Get sampling time.

        :return: dt (seconds)
        :rtype: float
        """
        return self.__dt

    @property
    def duration(self):
        """
        Get experiment duration.

        :return: duration (seconds)
        :rtype: float
        """
        return self.__duration

    @property
    def frequencies(self):
        """
        Get fundamental frequencies.

        :return: copy of the array of fundamental frequencies (hertz)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__frequencies)

    @property
    def t(self):
        """
        Get time.

        :return: copy of the time vector (seconds)
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__t)

    @property
    def x(self):
        """
        Get stimulation.

        :return: copy of the stimulation vector
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__x)

    @property
    def y(self):
        """
        Get response.

        :return: copy of the response vector
        :rtype: numpy.ndarray
        """
        return numpy.copy(self.__y)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name='qsa',
    version='1.0',
    author='Christophe Magnani',
    description='Linear and Quadratic Analysis in the Frequency Domain',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://openqsa.org',
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'matplotlib'],
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Science/Research',
        'Development Status :: 1 - Planning'
    ]
)
